<?php
  require(dirname(__FILE__) . DIRECTORY_SEPARATOR . 'udp.php');

  define('TTFTP_READ',  1);
  define('TTFTP_WRITE', 2);
  define('TTFTP_DATA',  3);
  define('TTFTP_ACK',   4);
  define('TTFTP_ERROR', 5);

  define('TTFTP_SUCESS',                 -1);
  define('TTFTP_ERROR_NOT_DEFINED',       0);
  define('TTFTP_ERROR_NOT_FOUND',         1);
  define('TTFTP_ERROR_ACCESS_VIOLATION',  2);
  define('TTFTP_ERROR_DISK_FULL',         3);
  define('TTFTP_ERROR_ILLEGAL_OPERATION', 4);
  define('TTFTP_ERROR_UNKNOWN_ID',        5);
  define('TTFTP_ERROR_FILE_EXISTS',       6);
  define('TTFTP_ERROR_UNKNOWN_USER',      7);

 /**
  * Templating Trivial File Transfer daemon
  *
  * TFTP is define in RFC 1350
  */
  class ttftp extends udp_daemon
  {
    var $max_data_length = 512;
    var $handlers = array();		// a chain of functions to call.

    function ttftp($bind_address='0.0.0.0', $bind_port=69)
    {
      $this->max_read = $this->max_data_length + 4; // 4 bytes for header
      parent::udp_daemon($bind_address, $bind_port);
    }

   /**
    * Add a handler to handle incoming read requests
    *
    * prototype:  my_handler($host, $filename, $mode, &$buffer);
    */
    function add_handler($handler)
    {
      $this->handlers[] = $handler;
    }

   /**
    * process an incoming packet
    *
    * called by udp::listen
    */
    function process_packet()
    {
      list(,$opcode) = unpack('n*', $this->client_buffer);
      switch($opcode)
      {
        case TTFTP_READ:  $this->process_read();                            break;
        case TTFTP_ACK:   $this->process_ack();                             break;
        case TTFTP_ERROR: $this->destroy_client();                          break;
        case TTFTP_WRITE: $this->send_error(TTFTP_ERROR_ACCESS_VIOLATION);  break;
        case TTFTP_DATA:  $this->send_error(TTFTP_ERROR_ACCESS_VIOLATION);  break;
        default:          $this->send_error(TTFTP_ERROR_ILLEGAL_OPERATION); break;
      }
    }

   /**
    * send a read request
    */
    function send_read($filename, $mode='octet')
    {
      $packet = pack('n', TTFTP_READ) . $filename . chr(0) . $mode . chr(0);
      return socket_sendto($this->client_socket, $packet, strlen($packet), 0, $this->client_address, $this->client_port);
    }

   /**
    * send a write request
    */
    function send_write($filename, $mode='octet')
    {
      $packet = pack('n', TTFTP_WRITE) . $filename . chr(0) . $mode . chr(0);
      return socket_sendto($this->client_socket, $packet, strlen($packet), 0, $this->client_address, $this->client_port);
    }

   /**
    * send a data packet
    */
    function send_data($block, $data=NULL)
    {
      $packet = pack('n', TTFTP_DATA) . pack('n', $block) . substr($data, 0, $this->max_data_length);
      return socket_sendto($this->client_socket, $packet, strlen($packet), 0, $this->client_address, $this->client_port);
    }

   /**
    * send an ack packet
    */
    function send_ack( $block)
    {
      $packet = pack('n', TTFTP_ACK) . pack('n', $block);
      return socket_sendto($this->client_socket, $packet, strlen($packet), 0, $this->client_address, $this->client_port);
    }

   /**
    * send an error packet
    */
    function send_error($code, $message=NULL)
    {
      if(is_null($message))
      {
        switch($code)
        {
          case TTFTP_ERROR_NOT_DEFINED:       $message = 'Not defined';                      break;
          case TTFTP_ERROR_NOT_FOUND:         $message = 'File not found';                   break;
          case TTFTP_ERROR_ACCESS_VIOLATION:  $message = 'Access violation';                 break;
          case TTFTP_ERROR_DISK_FULL:         $message = 'Disk full or allocation exceeded'; break;
          case TTFTP_ERROR_ILLEGAL_OPERATION: $message = 'Illegal TFTP operation';           break;
          case TTFTP_ERROR_UNKNOWN_ID:        $message = 'Unknown transfer ID';              break;
          case TTFTP_ERROR_FILE_EXISTS:       $message = 'File already exists';              break;
          case TTFTP_ERROR_UNKNOWN_USER:      $message = 'No such user';                     break;
        }
      }
      $packet = pack('n', TTFTP_ERROR) . pack('n', $code) . $message . chr(0);
      return socket_sendto($this->client_socket, $packet, strlen($packet), 0, $this->client_address, $this->client_port);
    }

   /**
    * process a read request
    */
    function process_read()
    {
      list($filename, $mode) = explode(chr(0), substr($this->client_buffer, 2));
      foreach($this->handlers as $handler)
      {
        if($handler($this->client_address, $filename, $mode, $this->client_buffer))
        {
          $parts = NULL;
          preg_match_all('/.{1,' . $this->max_data_length . '}/s', $this->client_buffer, $parts);
          $this->files[$this->client_index] = $parts[0];
          if(last_length($this->files[$this->client_index]) == $this->max_data_length)
            $this->files[$this->client_index][] = '';
          $this->send_data(1, substr($this->client_buffer, 0, $this->max_data_length));
          return true;
        }
      }
      $this->send_error(TTFTP_ERROR_NOT_FOUND);
      return false;
    }

   /**
    * process an ack (request for next chunk)
    */
    function process_ack()
    {
      list(,,$block) = unpack('n*', $this->client_buffer);
      if(isset($this->files[$this->client_index][$block]))
        $this->send_data($block + 1, $this->files[$this->client_index][$block]);
      else
        $this->destroy_client();
    }

   /**
    * Destroy a client
    *
    * @param integer $index of client
    */
    function destroy_client($index=NULL)
    {
      if(is_null($index)) $index = $this->client_index;
      unset($this->files[$index]);
      parent::destroy_client($index);
    }
  }

  function last_length($array)
  {
    return strlen($array[count($array)-1]);
  }
?>
